# =============================================================================
# Casanova HTTP Helpers
# =============================================================================
#
# HTTP Helpers working with optional urllib3 & certifi deps.
#
try:
    import urllib3
except ImportError:
    urllib3 = None

try:
    import certifi
except ImportError:
    certifi = None

from casanova.exceptions import NoHTTPSupport

urllib3_installed = urllib3 is not None
certifi_installed = certifi is not None

pool_manager = None

if urllib3_installed:
    manager_kwargs = {}

    if not certifi_installed:
        manager_kwargs["cert_reqs"] = "CERT_NONE"
    else:
        manager_kwargs["cert_reqs"] = "CERT_REQUIRED"
        manager_kwargs["ca_certs"] = certifi.where()

    pool_manager = urllib3.PoolManager(**manager_kwargs)


def request(url):
    if pool_manager is None:
        raise NoHTTPSupport(
            "casanova is not able to make http calls. please install urllib3 (and certifi if you want secure HTTPS)"
        )

    response = pool_manager.request("GET", url, preload_content=False)

    # Ref: https://github.com/urllib3/urllib3/issues/1305
    response._fp.isclosed = lambda: False

    return response



# =============================================================================
# Casanova Reader
# =============================================================================
#
# A fast but comfortable CSV reader based upon csv.reader to avoid dealing
# with csv.DictReader which is nice but very slow.
#
import csv
from os import PathLike
from collections import namedtuple
from collections.abc import Iterable
from itertools import chain
from io import IOBase, TextIOWrapper
from ebbe import without_last

from casanova.defaults import DEFAULTS
from casanova.headers import Headers
from casanova.namedrecord import is_tabular_record_class
from casanova.utils import (
    ensure_open,
    suppress_BOM,
    size_of_row_in_file,
    lines_without_null_bytes,
    rows_without_null_bytes,
    create_csv_aware_backwards_lines_iterator,
    ltpy311_csv_reader,
    looks_like_url,
    infer_delimiter_or_type,
)
from casanova.http import request
from casanova.exceptions import MissingColumnError, NoHeadersError

Multiplexer = namedtuple(
    "Multiplexer", ["column", "separator", "new_column"], defaults=["|", None]
)


class Reader(object):
    namespace = "casanova.reader"

    def __init__(
        self,
        input_file,
        no_headers=False,
        encoding="utf-8",
        dialect=None,
        quotechar=None,
        delimiter=None,
        prebuffer_bytes=None,
        total=None,
        multiplex=None,
        strip_null_bytes_on_read=None,
        reverse=False,
    ):
        # Resolving global defaults
        if prebuffer_bytes is None:
            prebuffer_bytes = DEFAULTS.prebuffer_bytes

        if strip_null_bytes_on_read is None:
            strip_null_bytes_on_read = DEFAULTS.strip_null_bytes_on_read

        if not isinstance(strip_null_bytes_on_read, bool):
            raise TypeError('expecting a boolean as "strip_null_bytes_on_read" kwarg')

        self.strip_null_bytes_on_read = strip_null_bytes_on_read

        if delimiter is None:
            data_format, inferred_delimiter = infer_delimiter_or_type(input_file)

            if data_format == "csv" and inferred_delimiter is not None:
                delimiter = inferred_delimiter

        # Detecting input type
        if isinstance(input_file, IOBase):
            input_type = "file"

        elif isinstance(input_file, (str, PathLike)):
            if isinstance(input_file, str) and looks_like_url(input_file):
                input_type = "url"
                input_file = request(input_file)
            else:
                input_type = "path"
                input_file = ensure_open(input_file, encoding=encoding)

        elif isinstance(input_file, Iterable):
            input_type = "iterable"
            input_file = iter(input_file)

        else:
            raise TypeError("expecting a file, a path or an iterable of rows")

        if multiplex is not None and not isinstance(multiplex, Multiplexer):
            raise TypeError("`multiplex` should be a casanova.Multiplexer")

        reader_kwargs = {}

        if dialect is not None:
            reader_kwargs["dialect"] = dialect
        if quotechar is not None:
            reader_kwargs["quotechar"] = quotechar
        if delimiter is not None:
            reader_kwargs["delimiter"] = delimiter

        self.input_type = input_type
        self.input_file = None
        self.backward_file = None

        if self.input_type == "iterable":
            if strip_null_bytes_on_read:
                self.reader = rows_without_null_bytes(input_file)
            else:
                self.reader = input_file

        else:
            self.input_file = input_file

            if self.input_type == "url":
                input_file = TextIOWrapper(
                    self.input_file, encoding=encoding, newline="", write_through=True
                )

            if strip_null_bytes_on_read:
                self.reader = csv.reader(
                    lines_without_null_bytes(input_file), **reader_kwargs
                )
            else:
                self.reader = ltpy311_csv_reader(input_file, **reader_kwargs)

        self.__buffered_rows = []
        self.was_completely_buffered = False
        self.total = total
        self.headers = None
        self.empty = False
        self.reverse = reverse
        self.no_headers = no_headers

        self.__no_headers_row_len = None

        # Reading headers
        if no_headers:
            try:
                self.__buffered_rows.append(next(self.reader))
            except StopIteration:
                self.empty = True
            else:
                self.__no_headers_row_len = len(self.__buffered_rows[0])
        else:
            try:
                fieldnames = next(self.reader)

                if fieldnames:
                    fieldnames[0] = suppress_BOM(fieldnames[0])

            except StopIteration:
                self.empty = True
            else:
                self.headers = Headers(fieldnames)

                try:
                    self.__buffered_rows.append(next(self.reader))
                except StopIteration:
                    self.empty = True

        # Reversing
        if reverse and not self.empty:
            if self.input_file is None or self.input_type == "url":
                raise NotImplementedError

            self.__buffered_rows.clear()
            (
                self.backward_file,
                backwards_lines_iterator,
            ) = create_csv_aware_backwards_lines_iterator(
                self.input_file,
                quotechar=quotechar,
                strip_null_bytes_on_read=strip_null_bytes_on_read,
            )

            self.reader = (
                ltpy311_csv_reader if not strip_null_bytes_on_read else csv.reader
            )(backwards_lines_iterator, **reader_kwargs)

            if not no_headers:
                self.reader = without_last(self.reader)

        # Multiplexing
        if multiplex is not None:
            if self.no_headers:
                multiplex_pos = multiplex.column

                if multiplex.new_column is not None:
                    raise TypeError(
                        "multiplexer cannot rename the column with no_headers=True"
                    )
            else:
                if multiplex.column not in self.headers:
                    raise MissingColumnError(multiplex.column)

                multiplex_pos = self.headers[multiplex.column]

            # New col
            if multiplex.new_column is not None:
                self.headers = Headers.rename(
                    self.headers, multiplex.column, multiplex.new_column
                )

            original_reader = self.reader
            already_buffered_rows = []

            if self.__buffered_rows:
                already_buffered_rows.append(self.__buffered_rows.pop())

            def multiplexing_reader():
                for row in chain(already_buffered_rows, original_reader):
                    cell = row[multiplex_pos]

                    if multiplex.separator not in cell:
                        yield row

                    else:
                        for value in cell.split(multiplex.separator):
                            copy = list(row)
                            copy[multiplex_pos] = value
                            yield copy

            self.reader = multiplexing_reader()

        # Prebuffering
        # NOTE: does not take into account probable first row of actual data
        # except the header one.
        if prebuffer_bytes is not None and self.total is None:
            if not isinstance(prebuffer_bytes, int) or prebuffer_bytes < 1:
                raise TypeError(
                    'expecting a positive integer as "prebuffer_bytes" kwarg'
                )

            buffered_bytes = 0

            while buffered_bytes < prebuffer_bytes:
                row = next(self.reader, None)

                if row is None:
                    self.was_completely_buffered = True
                    self.total = len(self.__buffered_rows)
                    break

                buffered_bytes += size_of_row_in_file(row)
                self.__buffered_rows.append(row)

        self.__buffered_rows.reverse()

        # Iteration state
        self.prelude_rows = None
        self.row_filter = None
        self.current_row_index = -1
        self.__iterator = self.__create_inner_rows_iterator()

    def __repr__(self):
        return "<%s>" % self.namespace

    @property
    def fieldnames(self):
        if self.headers is None:
            return None

        return self.headers.fieldnames

    @property
    def row_len(self):
        if self.__no_headers_row_len is not None:
            return self.__no_headers_row_len

        return len(self.headers)

    def __create_inner_rows_iterator(self):
        def chained():
            if self.prelude_rows is not None:
                for row in self.prelude_rows:
                    self.current_row_index += 1
                    yield row

            while self.__buffered_rows:
                self.current_row_index += 1
                yield self.__buffered_rows.pop()

            for row in self.reader:
                self.current_row_index += 1
                yield row

        if self.row_filter is None:
            yield from chained()

        for row in chained():
            if self.row_filter(self.current_row_index, row):
                yield row

    def rows(self):
        return self.__iterator

    def __next__(self):
        return next(self.__iterator)

    def __iter__(self):
        return self.rows()

    # NOTE: this function exists because it takes into
    # account rows skipped by resumers, which is
    # essential for threadsafe operations etc.
    def enumerate(self, start=0):
        for row in self.rows():
            yield self.current_row_index + start, row

    def wrap(self, row):
        return self.headers.wrap(row)

    # NOTE: the underlying cell iterator is protected so
    # it remains easy to override it and reimplement it
    # when inheriting (check the threadsafe_enricher)
    # for such an example.
    def __cells(self, column, with_rows=False):
        if not isinstance(column, (str, int)):
            raise TypeError("target column should be given as a name or a position")

        if not isinstance(column, int):
            if self.headers is None:
                raise NoHeadersError

            pos = self.headers.get(column)

            if pos is None:
                raise MissingColumnError(column)
        else:
            if column >= self.row_len:
                raise MissingColumnError

            pos = column

        if with_rows:

            def iterator():
                for row in self.rows():
                    yield row, row[pos]

        else:

            def iterator():
                for row in self.rows():
                    yield row[pos]

        return iterator()

    def cells(self, column, *, with_rows=False):
        return self.__cells(column, with_rows=with_rows)

    def enumerate_cells(self, column, start=0, *, with_rows=False):
        if with_rows:
            for row, value in self.__cells(column, with_rows=True):
                yield self.current_row_index + start, row, value
        else:
            for value in self.__cells(column):
                yield self.current_row_index + start, value

    # NOTE: the underlying records iterator is protected so
    # it remains easy to override it and reimplement it
    # when inheriting (check the threadsafe_enricher)
    # for such an example.
    def __records(self, *shape, with_rows=False, ignore_headers=False):
        if len(shape) < 1:
            raise TypeError("shape of record was not given")

        if len(shape) == 1 and getattr(shape[0], "is_namedrecord", False):
            if (
                not ignore_headers
                and self.fieldnames is not None
                and self.fieldnames != shape[0].fieldnames
            ):
                raise TypeError("file headers are not aligned with the record ones")

            project = lambda r: shape[0](*r)

        elif len(shape) == 1 and is_tabular_record_class(shape[0]):
            if (
                not ignore_headers
                and self.fieldnames is not None
                and self.fieldnames != shape[0].fieldnames()
            ):
                raise TypeError("file headers are not aligned with the record ones")

            project = lambda r: shape[0].parse(r)
        elif self.no_headers:
            project = Headers.flat_project_no_headers(self.row_len, *shape)
        else:
            project = self.headers.flat_project(*shape)

        if with_rows:

            def iterator():
                for row in self.rows():
                    yield row, project(row)

        else:

            def iterator():
                for row in self.rows():
                    yield project(row)

        return iterator()

    def records(self, *shape, with_rows=False, ignore_headers=False):
        return self.__records(
            *shape, with_rows=with_rows, ignore_headers=ignore_headers
        )

    def enumerate_records(self, *shape, start=0, with_rows=False, ignore_headers=False):
        if with_rows:
            for row, record in self.__records(
                *shape, with_rows=True, ignore_headers=ignore_headers
            ):
                yield self.current_row_index + start, row, record
        else:
            for record in self.__records(*shape, ignore_headers=ignore_headers):
                yield self.current_row_index + start, record

    def close(self):
        if self.input_type == "url":
            self.input_file.release_conn()

        if self.input_file is not None and hasattr(self.input_file, "close"):
            self.input_file.close()

        if self.backward_file is not None:
            self.backward_file.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @classmethod
    def count(cls, input_file, max_rows=None, **kwargs):
        assert max_rows is None or max_rows > 0, (
            "%s.count: expected max_rows to be `None` or > 0." % cls.namespace
        )

        n = 0

        with cls(input_file, **kwargs) as reader:
            for _ in reader:
                n += 1

                if max_rows is not None and n > max_rows:
                    return None

        return n



# =============================================================================
# Casanova Resuming Strategies
# =============================================================================
#
# A collection of process resuming strategies acknowledged by casanova
# enrichers.
#
from threading import Lock
from os.path import isfile, getsize
from collections import deque, namedtuple

from casanova.reader import Reader
from casanova.reverse_reader import ReverseReader
from casanova.exceptions import (
    ResumeError,
    NotResumableError,
    MissingColumnError,
    CorruptedIndexColumnError,
)
from casanova.contiguous_range_set import ContiguousRangeSet


class Resumer(object):
    def __init__(self, path, listener=None, encoding="utf-8"):
        self.path = path
        self.encoding = encoding
        self.output_file = None
        self.lock = Lock()
        self.popped = False
        self.buffer = deque()

        self.listener = None

        if listener is not None:
            self.set_listener(listener)

    def set_listener(self, listener):
        if not callable(listener):
            raise TypeError("listener should be callable")

        self.listener = listener

    def can_resume(self):
        return isfile(self.path) and getsize(self.path) > 0

    def open(self, mode="a", newline=""):
        return open(self.path, mode=mode, encoding=self.encoding, newline=newline)

    def open_output_file(self, **kwargs):
        if self.output_file is not None:
            raise ResumeError("output file is already opened")

        mode = "a+" if self.can_resume() else "w"

        self.output_file = self.open(mode=mode, **kwargs)
        return self.output_file

    def emit(self, event, payload):
        if self.listener is None:
            return

        with self.lock:
            self.listener(event, payload)

    def get_insights_from_output(self, _):
        raise NotImplementedError

    def filter_row(self, i, row):
        result = self.filter(i, row)

        if not result:
            self.emit("input.row.filter", row)

        return result

    def get_state(self):
        raise NotImplementedError

    def pop_state(self):
        if not self.popped:
            self.popped = True
            return self.get_state()

        return None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def flush(self) -> None:
        if self.output_file is not None:
            self.output_file.flush()

    def close(self):
        if self.output_file is not None:
            self.output_file.close()
            self.output_file = None

    def __repr__(self):
        return "<{name} path={path!r} can_resume={can_resume!r}>".format(
            name=self.__class__.__name__, path=self.path, can_resume=self.can_resume()
        )

    def already_done_count(self):
        raise NotImplementedError

    def __iter__(self):
        if hasattr(self, "filter"):
            raise NotImplementedError

        while self.buffer:
            yield self.buffer.popleft()


class RowCountResumer(Resumer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.row_count = 0

    def get_insights_from_output(self, _, **reader_kwargs):
        self.row_count = 0

        with self.open(mode="r") as f:
            reader = Reader(f, **reader_kwargs)

            count = 0

            for row in reader:
                self.emit("output.row.read", row)
                count += 1

        self.row_count = count

    def resume(self, enricher):
        i = 0
        iterator = iter(enricher)

        while i < self.row_count:
            row = next(iterator)
            self.emit("input.row.filter", row)
            i += 1

    def already_done_count(self):
        return self.row_count


class ThreadSafeResumer(Resumer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.already_done = ContiguousRangeSet()

    def get_insights_from_output(self, enricher, **reader_kwarg):
        self.already_done = ContiguousRangeSet()

        with self.open(mode="r") as f:
            reader = Reader(f, **reader_kwarg)

            pos = reader.headers.get(enricher.index_column)

            if pos is None:
                raise MissingColumnError(enricher.index_column)

            for row in reader:
                self.emit("output.row.read", row)

                try:
                    current_index = int(row[pos])
                except ValueError:
                    raise CorruptedIndexColumnError

                self.already_done.add(current_index)

    def filter(self, i, row):
        return not self.already_done.stateful_contains(i)

    def already_done_count(self):
        return len(self.already_done)


BatchResumerContext = namedtuple(
    "BatchResumerContext", ["last_cursor", "values_to_skip"]
)


class BatchResumer(Resumer):
    def __init__(self, path, value_column, **kwargs):
        super().__init__(path, **kwargs)
        self.last_batch = None
        self.value_column = value_column
        self.value_pos = None
        self.last_cursor = None
        self.values_to_skip = None

    def get_insights_from_output(self, enricher, **reader_kwargs):
        self.last_batch = ReverseReader.last_batch(
            self.path,
            batch_value=self.value_column,
            batch_cursor=enricher.cursor_column,
            end_symbol=enricher.end_symbol,
            **reader_kwargs
        )
        self.value_pos = enricher.output_headers[self.value_column]
        self.last_cursor = None
        self.values_to_skip = None

    def get_state(self):
        return BatchResumerContext(self.last_cursor, self.values_to_skip)

    def resume(self, enricher):
        last_batch = self.last_batch

        if last_batch is None:
            return

        iterator = iter(enricher)

        while True:
            row = next(iterator, None)

            if row is None:
                raise NotResumableError

            self.emit("input.row.filter", row)

            value = row[self.value_pos]

            # We haven't reached our batch yet
            if value != last_batch.value:
                continue

            # Last batch was completely finished
            elif last_batch.finished:
                break

            # Here we need to record additional information
            self.last_cursor = last_batch.cursor
            self.values_to_skip = set(row[self.value_pos] for row in last_batch.rows)
            self.buffer.append(row)

            break


class LastCellResumer(Resumer):
    def __init__(self, path, value_column, **kwargs):
        super().__init__(path, **kwargs)
        self.last_cell = None
        self.value_column = value_column

    def get_insights_from_output(self, _, **reader_kwargs):
        self.last_cell = ReverseReader.last_cell(
            self.path, column=self.value_column, **reader_kwargs
        )

    def get_state(self):
        return self.last_cell


class LastCellComparisonResumer(LastCellResumer):
    """
    Warning : this resumer will not work as desired if the column read contains duplicate values.
    """

    def resume(self, enricher):
        for row in enricher:
            self.emit("input.row.filter", row)

            if row[self.value_column] == self.last_cell:
                break



from typing import Optional

import os
import sys
import shlex
import shutil
import platform
import multiprocessing
from argparse import ArgumentParser, HelpFormatter, ArgumentTypeError
from functools import partial

from casanova.defaults import set_defaults
from casanova.utils import ensure_open, LT_PY311
from casanova.cli import (
    map_action,
    flatmap_action,
    filter_action,
    map_reduce_action,
    groupby_action,
    reverse_action,
)


def acquire_cross_platform_stdout():
    # As per #254: stdout need to be wrapped so that windows get a correct csv
    # stream output
    if "windows" in platform.system().lower():
        return open(
            sys.__stdout__.fileno(),
            mode=sys.__stdout__.mode,
            buffering=1,
            encoding=sys.__stdout__.encoding,
            errors=sys.__stdout__.errors,
            newline="",
            closefd=False,
        )

    return sys.stdout


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def die(*args):
    print_err(*args)
    sys.exit(1)


class SortingHelpFormatter(HelpFormatter):
    def add_arguments(self, actions) -> None:
        actions = sorted(
            actions, key=lambda a: tuple(s.lower() for s in a.option_strings)
        )
        return super().add_arguments(actions)


def custom_formatter(prog):
    terminal_size = shutil.get_terminal_size()
    return SortingHelpFormatter(prog, width=terminal_size.columns, max_help_position=32)


VALID_ARG_NAMES = {"index", "row", "headers", "fieldnames", "cell", "cells"}


class ArgsType:
    def __call__(self, string):
        args = []

        if not string.strip():
            return args

        for s in string.split(","):
            arg_name = s.strip().lower()

            if arg_name not in VALID_ARG_NAMES:
                raise ArgumentTypeError(
                    "%s is not a valid arg name. Must be one of: %s"
                    % (s, VALID_ARG_NAMES)
                )

            args.append(arg_name)

        return args


class SpliterType:
    def __init__(self, splitchar=","):
        self.splitchar = splitchar

    def __call__(self, string):
        return string.split(self.splitchar)


COMMON_ARGUMENTS = [
    (
        ("-d", "--delimiter"),
        {"help": 'CSV delimiter to use. Defaults to ",".', "default": ","},
    ),
    (
        ("-o", "--output"),
        {
            "help": "Path to the output file. Will default to stdout and will consider `-` as stdout."
        },
    ),
]

MP_ARGUMENTS = [
    (
        ("-p", "--processes"),
        {
            "help": "Number of processes to use. Defaults to 1.",
            "default": 1,
            "type": int,
        },
    ),
    (
        (
            "-c",
            "--chunk-size",
        ),
        {
            "help": "Multiprocessing chunk size. Defaults to 1.",
            "default": 1,
            "type": int,
        },
    ),
    (
        ("-u", "--unordered"),
        {
            "help": "Whether you allow the result to be in arbitrary order when using multiple processes. Defaults to no.",
            "action": "store_true",
        },
    ),
    (
        ("-I", "--init"),
        {
            "help": "Code to execute once before starting to iterate over file. Useful to setup global variables used in evaluated code later. Can be given multiple times.",
            "action": "append",
            "default": [],
        },
    ),
    (
        ("-B", "--before"),
        {
            "help": "Code to execute before each evaluation of code for a row in the CSV file. Useful to update variables before returning something. Can be given multiple times.",
            "action": "append",
            "default": [],
        },
    ),
    (
        ("-A", "--after"),
        {
            "help": "Code to execute after each evaluation of code for a row in the CSV file. Useful to update variables after having returned something. Can be given multiple times.",
            "action": "append",
            "default": [],
        },
    ),
    (
        ("-m", "--module"),
        {
            "help": "If set, given code will be interpreted as a python module to import and a function name taking the current index and row.",
            "action": "store_true",
        },
    ),
    (
        ("-a", "--args"),
        {
            "help": 'List of arguments to pass to the function when using -m/--module. Defaults to "row".',
            "default": ["row"],
            "type": ArgsType(),
        },
    ),
    (
        ("-i", "--ignore-errors"),
        {
            "help": "If set, evaluation raising an error will be considered as returning None instead of raising.",
            "action": "store_true",
        },
    ),
    (
        ("-s", "--select"),
        {
            "help": 'Use to select columns. First selected column value will be forwared as "cell" and selected column values as "cells".'
        },
    ),
]

SERIALIZATION_ARGUMENTS = [
    (
        ("--plural-separator",),
        {
            "help": 'Character to use to join lists and sets together in a single cell. Defaults to "|". If you need to emit multiple rows instead, consider using flatmap.',
            "default": "|",
        },
    ),
    (
        ("--none-value",),
        {
            "help": "String used to serialize None values. Defaults to an empty string.",
            "default": "",
        },
    ),
    (
        ("--true-value",),
        {
            "help": 'String used to serialize True values. Defaults to "true".',
            "default": "true",
        },
    ),
    (
        ("--false-value",),
        {
            "help": 'String used to serialize False values. Defaults to "false".',
            "default": "false",
        },
    ),
]

FORMAT_ARGUMENTS = [
    (
        ("--json",),
        {"help": "Whether to format the output as json.", "action": "store_true"},
    ),
    (
        ("--pretty",),
        {
            "help": "Whether to prettify the output, e.g. indent the json file.",
            "action": "store_true",
        },
    ),
    (
        ("--csv",),
        {"help": "Whether to format the output as csv.", "action": "store_true"},
    ),
]


def add_arguments(parser: ArgumentParser, arguments):
    for args, kwargs in arguments:
        parser.add_argument(*args, **kwargs)


add_common_arguments = partial(add_arguments, arguments=COMMON_ARGUMENTS)
add_mp_arguments = partial(add_arguments, arguments=MP_ARGUMENTS)
add_serialization_arguments = partial(add_arguments, arguments=SERIALIZATION_ARGUMENTS)
add_format_arguments = partial(add_arguments, arguments=FORMAT_ARGUMENTS)


def build_commands():
    parser = ArgumentParser(
        "casanova",
        description="Casanova command line utilities such as mapping, filtering, reducing columns of a given CSV files.",
        formatter_class=custom_formatter,
    )
    add_common_arguments(parser)

    subparsers = parser.add_subparsers(dest="action", help="Command to execute.")

    map_parser = subparsers.add_parser("map", formatter_class=custom_formatter)
    add_common_arguments(map_parser)
    add_mp_arguments(map_parser)
    add_serialization_arguments(map_parser)
    map_parser.add_argument(
        "new_column",
        help="Name of the new column to create & containing the result of the evaluated code.",
    )
    map_parser.add_argument(
        "code", help="Python code to evaluate for each row of the CSV file."
    )
    map_parser.add_argument(
        "file",
        help="CSV file to map. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )

    flatmap_parser = subparsers.add_parser("flatmap", formatter_class=custom_formatter)
    add_common_arguments(flatmap_parser)
    add_mp_arguments(flatmap_parser)
    add_serialization_arguments(flatmap_parser)
    flatmap_parser.add_argument(
        "new_column",
        help="Name of the new column to create & containing the result of the evaluated code.",
    )
    flatmap_parser.add_argument(
        "code", help="Python code to evaluate for each row of the CSV file."
    )
    flatmap_parser.add_argument(
        "file",
        help="CSV file to flatmap. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )

    filter_parser = subparsers.add_parser("filter", formatter_class=custom_formatter)
    add_common_arguments(filter_parser)
    add_mp_arguments(filter_parser)
    filter_parser.add_argument(
        "code", help="Python code to evaluate for each row of the CSV file."
    )
    filter_parser.add_argument(
        "file",
        help="CSV file to filter. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )

    map_reduce_parser = subparsers.add_parser(
        "map-reduce", formatter_class=custom_formatter
    )
    add_common_arguments(map_reduce_parser)
    add_mp_arguments(map_reduce_parser)
    add_serialization_arguments(map_reduce_parser)
    add_format_arguments(map_reduce_parser)
    map_reduce_parser.add_argument(
        "code", help="Python code to evaluate for each row of the CSV file."
    )
    map_reduce_parser.add_argument(
        "accumulator",
        help="Python code that will be evaluated to perform the accumulation towards a final value.",
    )
    map_reduce_parser.add_argument(
        "file",
        help="CSV file to map-reduce. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )
    map_reduce_parser.add_argument(
        "-V",
        "--init-value",
        help="Python code to evaluate to initialize the accumulator's value. If not given, the initial value will be the first map result.",
    )

    groupby_parser = subparsers.add_parser("groupby", formatter_class=custom_formatter)
    add_common_arguments(groupby_parser)
    add_mp_arguments(groupby_parser)
    add_serialization_arguments(groupby_parser)
    groupby_parser.add_argument(
        "code", help="Python code to evaluate to group each row of the CSV file."
    )
    groupby_parser.add_argument(
        "aggregator",
        help="Python code that will be evaluated to perform the aggregation of each yielded group of rows.",
    )
    groupby_parser.add_argument(
        "file",
        help="CSV file to group. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )
    groupby_parser.add_argument(
        "-f",
        "--fieldnames",
        help="Output CSV file fieldnames. Useful when emitting sequences without keys (e.g. lists, tuples etc.).",
        type=SpliterType(),
    )

    reverse_parser = subparsers.add_parser("reverse", formatter_class=custom_formatter)
    add_common_arguments(reverse_parser)
    reverse_parser.add_argument(
        "file",
        help="CSV file to read in reverse. Can be gzip-compressed, and can also be a URL. Will consider `-` as stdin.",
    )

    commands = {
        "map": (map_parser, map_action),
        "flatmap": (flatmap_parser, flatmap_action),
        "filter": (filter_parser, filter_action),
        "map-reduce": (map_reduce_parser, map_reduce_action),
        "groupby": (groupby_parser, groupby_action),
        "reverse": (reverse_parser, reverse_action),
    }

    return parser, commands


CASANOVA_PARSER, CASANOVA_COMMANDS = build_commands()


def main(arguments_override: Optional[str] = None):
    cli_args = CASANOVA_PARSER.parse_args(
        shlex.split(arguments_override) if arguments_override is not None else None
    )

    if cli_args.action is None:
        CASANOVA_PARSER.print_help()
        sys.exit(0)

    _, action = CASANOVA_COMMANDS[cli_args.action]

    # Validating
    args_flag = getattr(cli_args, "args", [])

    if "cell" in args_flag or "cells" in args_flag:
        if not getattr(cli_args, "select", None):
            die('Cannot use "cell" or "cells" in --args without providing -s/--select!')

    # Stdin fallback
    if getattr(cli_args, "file", None) == "-":
        cli_args.file = sys.stdin

    # Dealing with output stream
    if cli_args.output is None or cli_args.output == "-":
        action(cli_args, acquire_cross_platform_stdout())
    else:
        with ensure_open(
            cli_args.output, "w", encoding="utf-8", newline=""
        ) as output_file:
            action(cli_args, output_file)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    multiprocessing.set_start_method("spawn")

    if LT_PY311:
        set_defaults(strip_null_bytes_on_read=True, strip_null_bytes_on_write=True)

    try:
        main()
    except (KeyboardInterrupt, BrokenPipeError):
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)



# =============================================================================
# Casanova Headers
# =============================================================================
#
# Utility class representing a CSV file's headers
#
import re
from ebbe import with_next
from collections import namedtuple, defaultdict
from collections.abc import Iterable
from functools import wraps

from casanova.exceptions import (
    InvalidSelectionError,
    MissingColumnError,
    NthNamedColumnOutOfRangeError,
    UnknownNamedColumnError,
    ColumnOutOfRangeError,
)

PROJECTION_SHAPE_TYPES = (int, str, tuple, list, dict)


class Selection(object):
    def __init__(self, inverted=False):
        self.groups = []
        self.inverted = inverted

    def add(self, group):
        self.groups.append(group)

    def is_suitable_without_headers(self) -> bool:
        for group in self.groups:
            if isinstance(group, SingleColumn):
                if not isinstance(group.key, int):
                    return False

            elif isinstance(group, IndexedColumn):
                return False

            elif isinstance(group, ColumnRange):
                # NOTE: start and end cannot be mixed, so it does not
                # make sense to also test end
                if not isinstance(group.start, int):
                    return False

            else:
                return False

        return True

    def __iter__(self):
        yield from self.groups


SingleColumn = namedtuple("SingleColumn", ["key"])
ColumnRange = namedtuple("ColumnRange", ["start", "end"])
IndexedColumn = namedtuple("IndexedColumn", ["key", "index"])

INDEXED_HEADER_RE = re.compile(r"^.+\[(\d+)\]$")
INDEX_REPLACER_RE = re.compile(r"\[\d+\]$")


def redirect_errors_as_invalid_selection(fn):
    @wraps(fn)
    def wrapped(self, selection):
        try:
            return fn(self, selection)
        except MissingColumnError as e:
            raise InvalidSelectionError(selection=selection, reason=e)

    return wrapped


def parse_key(key):
    if not key:
        return None

    if key.isdigit():
        return int(key)

    return INDEX_REPLACER_RE.sub("", key)


def parse_selection(string):
    """
    From xsv:

    Select one column by name:
        * name

    Select one column by index (1-based):
        * 2

    Select the first and fourth columns:
        * 1,4

    Select the first 4 columns (by index and by name):
        * 1-4
        * Header1-Header4

    Ignore the first 2 columns (by range and by omission):
        * 3-
        * '!1-2'

    Select the third column named 'Foo':
        * 'Foo[2]'

    Re-order and duplicate columns arbitrarily:
        * 3-1,Header3-Header1,Header1,Foo[2],Header1

    Quote column names that conflict with selector syntax:
        * '"Date - Opening","Date - Actual Closing"'
    """
    inverted = False

    if string.startswith("!"):
        inverted = True
        string = string[1:]

    selection = Selection(inverted=inverted)

    def tokens():
        acc = ""
        current_escapechar = None
        escaping = False

        for c in string:
            if c == "\\":
                escaping = True
                continue

            if escaping:
                escaping = False
                acc += c
                continue

            if c == current_escapechar:
                current_escapechar = None
                continue

            if c == "'" or c == '"':
                current_escapechar = c
                continue

            if current_escapechar is None:
                if c == ",":
                    yield (acc, ",")
                    acc = ""
                    continue

                elif c == "-":
                    yield (acc, "-")
                    acc = ""
                    continue

            acc += c

        if acc:
            yield (acc, None)

    def combined_tokens():
        init = True
        skip_next = False

        for (
            (token, sep),
            next_item,
        ) in with_next(tokens()):
            if skip_next:
                skip_next = False
                continue

            if not init and inverted:
                raise InvalidSelectionError(
                    reason=TypeError("invalid-inverted-selection"),
                    selection=string,
                )

            key = parse_key(token)

            init = False

            if sep == "-":
                skip_next = True

                end_key = parse_key(next_item[0]) if next_item is not None else None

                if end_key is not None and type(key) is not type(end_key):
                    raise InvalidSelectionError(
                        reason=TypeError("mixed-range"), selection=string
                    )

                # NOTE: ranges are 1-based in xsv
                if isinstance(key, int):
                    key -= 1

                    if end_key is not None:
                        end_key -= 1

                    if key < 0 or (end_key is not None and end_key < 0):
                        raise InvalidSelectionError(
                            reason=TypeError("negative-index"), selection=string
                        )

                if end_key is not None and key == end_key:
                    yield SingleColumn(key=key)
                    continue

                yield ColumnRange(start=key, end=end_key)
                continue

            index_match = INDEXED_HEADER_RE.match(token)

            if index_match:
                if isinstance(key, int):
                    raise InvalidSelectionError(
                        reason=TypeError("invalid-indexation"), selection=string
                    )

                yield IndexedColumn(key=key, index=int(index_match.group(1)))
                continue

            if isinstance(key, int):
                key -= 1

                if key < 0:
                    raise InvalidSelectionError(
                        reason=TypeError("negative-index"), selection=string
                    )

            yield SingleColumn(key=key)

    for group in combined_tokens():
        selection.add(group)

    return selection


def walk_shape(node, headers):
    if isinstance(node, (int, str)):
        return headers[node]

    if isinstance(node, list):
        return [walk_shape(n, headers) for n in node]

    if isinstance(node, tuple):
        return tuple(walk_shape(n, headers) for n in node)

    if isinstance(node, dict):
        o = {}

        for k, v in node.items():
            o[k] = walk_shape(v, headers)

        return o

    raise NotImplementedError


def walk_row(node, row):
    if isinstance(node, (int)):
        return row[node]

    if isinstance(node, list):
        return [walk_row(n, row) for n in node]

    if isinstance(node, tuple):
        return tuple(walk_row(n, row) for n in node)

    if isinstance(node, dict):
        o = {}

        for k, v in node.items():
            o[k] = walk_row(v, row)

        return o

    raise NotImplementedError


class RowWrapper(object):
    __slots__ = ("__headers", "__row")

    def __init__(self, headers, row):
        self.__headers = headers
        self.__row = row

    def _replace(self, row):
        self.__row = row

    def __getitem__(self, key):
        return self.__row[self.__headers[key]]

    def get(self, key, default=None, index=None):
        idx = self.__headers.get(key, index=index)

        if idx is None:
            return default

        return self.__row[idx]

    def __contains__(self, key):
        return key in self.__headers

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __iter__(self):
        yield from self.__row

    def __len__(self):
        return len(self.__row)

    def cells(self):
        yield from zip(self.__headers.fieldnames, self.__row)


class Headers(object):
    def __init__(self, fieldnames):
        self.fieldnames = list(fieldnames)

        self.__mapping = defaultdict(list)
        self.__flat_mapping = {}
        self.__wrapper = RowWrapper(self, fieldnames)

        for i, h in enumerate(self.fieldnames):
            self.__mapping[h].append(i)
            self.__flat_mapping[h] = i

    def __eq__(self, other):
        return self.fieldnames == other.fieldnames

    def __len__(self):
        return len(self.fieldnames)

    def __getitem__(self, key):
        if isinstance(key, int):
            if key >= len(self):
                raise ColumnOutOfRangeError(key)

            return key

        if isinstance(key, tuple):
            if len(key) != 2:
                raise TypeError("expecting a str, a int or a (str, int) tuple")

            indices = self.__mapping.get(key[0])

            if indices is None:
                raise UnknownNamedColumnError(key[0])

            try:
                return indices[key[1]]
            except IndexError:
                raise NthNamedColumnOutOfRangeError(key[0], key[1])

        indices = self.__mapping.get(key)

        if indices is None:
            raise UnknownNamedColumnError(key)

        assert len(indices) > 0

        return indices[0]

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __contains__(self, key):
        if isinstance(key, int):
            return key < len(self)

        if isinstance(key, tuple):
            if len(key) != 2:
                raise TypeError("expecting a str, a int or a (str, int) tuple")

            indices = self.__mapping.get(key[0])

            if indices is None:
                return False

            return key[1] < len(indices)

        return key in self.__mapping

    def __iter__(self):
        yield from self.fieldnames

    def nth(self, index):
        try:
            return self.fieldnames[index]
        except IndexError:
            raise ColumnOutOfRangeError(index)

    def get(self, key, default=None, index=None):
        if isinstance(key, tuple):
            if len(key) != 2:
                raise TypeError("expecting a str, a int or a (str, int) tuple")

            if index is not None:
                raise TypeError(
                    "key cannot be a (name, index) tuple if index kwarg is not None"
                )

            index = key[1]
            key = key[0]

        if isinstance(key, int):
            if index is not None:
                raise TypeError("it does not make sense to get an nth index")

            if key >= len(self):
                return default

            return key

        indices = self.__mapping.get(key)

        if indices is None:
            return default

        assert len(indices) > 0

        if index is not None:
            return indices[index]

        return indices[0]

    @redirect_errors_as_invalid_selection
    def select(self, selection):
        indices = []

        if not isinstance(selection, (str, Selection)):
            if not isinstance(selection, Iterable):
                raise TypeError(
                    "invalid selection. expecting str, Selection or iterable"
                )

            for key in selection:
                indices.append(self[key])

            return indices

        parsed_selection = (
            parse_selection(selection)
            if not isinstance(selection, Selection)
            else selection
        )

        for group in parsed_selection:
            if isinstance(group, SingleColumn):
                key = self[group.key]

                if parsed_selection.inverted:
                    for i in range(len(self)):
                        if i == key:
                            continue

                        indices.append(i)
                else:
                    indices.append(key)

            elif isinstance(group, IndexedColumn):
                key = self[group.key, group.index]

                if parsed_selection.inverted:
                    for i in range(len(self)):
                        if i == key:
                            continue

                        indices.append(i)
                else:
                    indices.append(key)

            elif isinstance(group, ColumnRange):
                # NOTE: ranges are all inclusive in xsv dsl
                start = self[group.start]
                end = self[group.end] if group.end is not None else None

                target_range: range

                if end is not None:
                    if start > end:
                        target_range = range(start, end - 1, -1)
                    else:
                        target_range = range(start, end + 1)
                else:
                    target_range = range(start, len(self))

                if parsed_selection.inverted:
                    for i in range(len(self)):
                        if i in target_range:
                            continue

                        indices.append(i)
                else:
                    indices.extend(list(target_range))

            else:
                raise NotImplementedError(
                    "selection implementation is erroneously not exhaustive"
                )

        return indices

    def project(self, shape):
        if not isinstance(shape, PROJECTION_SHAPE_TYPES):
            raise TypeError("invalid projection shape")

        indexed_shape = walk_shape(shape, self)

        def projection(row):
            return walk_row(indexed_shape, row)

        return projection

    def flat_project(self, *args):
        if len(args) < 1:
            raise TypeError("not enough arguments")

        if len(args) > 1:
            shape = tuple(args)
        else:
            shape = args[0]

        if not isinstance(shape, PROJECTION_SHAPE_TYPES):
            raise TypeError("invalid projection shape")

        def select_one(target):
            if isinstance(target, int):
                return target

            result = self.select(target)

            if len(result) > 1:
                raise TypeError(
                    "projection shape includes a selection returning more than one column"
                )

            return result[0]

        if isinstance(shape, (int, str)):
            idx = select_one(shape)

            def projection(row):
                return row[idx]

        elif isinstance(shape, tuple):
            indices = [select_one(item) for item in shape]

            def projection(row):
                return tuple(row[i] for i in indices)

        elif isinstance(shape, list):
            indices = [select_one(item) for item in shape]

            def projection(row):
                return [row[i] for i in indices]

        elif isinstance(shape, dict):
            indices = {k: select_one(v) for k, v in shape.items()}

            def projection(row):
                return {k: row[v] for k, v in indices.items()}

        else:
            raise NotImplementedError

        return projection

    def wrap(self, row, transient=False):
        if len(row) != len(self.fieldnames):
            raise TypeError("len mismatch for row and headers")

        if transient:
            self.__wrapper._replace(row)
            return self.__wrapper

        return RowWrapper(self, row)

    def __repr__(self):
        class_name = self.__class__.__name__

        representation = "<" + class_name

        for i, h in enumerate(self):
            if h.isidentifier():
                representation += " %s=%s" % (h, i)
            else:
                representation += ' "%s"=%s' % (h, i)

        representation += ">"

        return representation

    @classmethod
    def select_no_headers(cls, count, selection):
        if isinstance(selection, str):
            parsed_selection = parse_selection(selection)

            if not parsed_selection.is_suitable_without_headers():
                raise InvalidSelectionError(
                    selection=selection,
                    reason=TypeError("irrelevant-naming-no-headers"),
                )

            selection = parsed_selection

        headers = cls(range(count))

        return headers.select(selection)

    @classmethod
    def flat_project_no_headers(cls, count, *shape):
        headers = cls(range(count))

        return headers.flat_project(*shape)

    @classmethod
    def rename(cls, headers: "Headers", old_name: str, new_name: str):
        new_fieldnames = []

        for f in headers.fieldnames:
            new_fieldnames.append(f if f != old_name else new_name)

        return cls(new_fieldnames)



# =============================================================================
# Casanova Global Defaults
# =============================================================================
#
# Global mutable defaults used by casanova classes.
#
from contextlib import contextmanager


class CasanovaDefaults(object):
    __slots__ = [
        "prebuffer_bytes",
        "strip_null_bytes_on_read",
        "strip_null_bytes_on_write",
    ]

    def __init__(self):
        self.prebuffer_bytes = None
        self.strip_null_bytes_on_read = False
        self.strip_null_bytes_on_write = False

    def save(self):
        return {n: getattr(self, n) for n in self.__slots__}

    def load(self, o):
        for n, v in o.items():
            setattr(self, n, v)


DEFAULTS = CasanovaDefaults()
NOT_GIVEN = object()


def set_defaults(
    prebuffer_bytes=NOT_GIVEN,
    strip_null_bytes_on_read=NOT_GIVEN,
    strip_null_bytes_on_write=NOT_GIVEN,
    plural_separator=NOT_GIVEN,
    none_value=NOT_GIVEN,
    true_value=NOT_GIVEN,
    false_value=NOT_GIVEN,
):
    global DEFAULTS

    if prebuffer_bytes is not NOT_GIVEN:
        if prebuffer_bytes is not None and (
            not isinstance(prebuffer_bytes, int) or prebuffer_bytes < 1
        ):
            raise TypeError("prebuffer_bytes should be None or a positive integer")

        DEFAULTS.prebuffer_bytes = prebuffer_bytes

    if strip_null_bytes_on_read is not NOT_GIVEN:
        if not isinstance(strip_null_bytes_on_read, bool):
            raise TypeError("strip_null_bytes_on_read should be a boolean")

        DEFAULTS.strip_null_bytes_on_read = strip_null_bytes_on_read

    if strip_null_bytes_on_write is not NOT_GIVEN:
        if not isinstance(strip_null_bytes_on_write, bool):
            raise TypeError("strip_null_bytes_on_write should be a boolean")

        DEFAULTS.strip_null_bytes_on_write = strip_null_bytes_on_write


@contextmanager
def temporary_defaults(**kwargs):
    try:
        original = DEFAULTS.save()
        set_defaults(**kwargs)
        yield
    finally:
        DEFAULTS.load(original)



# =============================================================================
# Casanova Named Record Helper
# =============================================================================
#
# CSV-aware improvement over python's namedtuple.
#
import sys
import json
from typing import Optional, Iterable, Union, List

if sys.version_info[:2] >= (3, 10):
    from typing import get_origin, get_args
else:
    from typing_extensions import get_origin, get_args

from collections import OrderedDict, namedtuple
from dataclasses import fields, field, Field

from casanova.serialization import CSVSerializer

DEFAULT = 0
JSON = 1


# NOTE: boolean & plural are just indicative and don't serve any purpose
# anymore but to be additional metadata that could be useful later on
# NOTE: json could also become indicative only at one point
def namedrecord(
    name: str,
    fields: Iterable[str],
    boolean=None,
    plural=None,
    json: Optional[Iterable[str]] = None,
    defaults: Optional[Iterable] = None,
    plural_separator: Optional[str] = None,
    none_value: Optional[str] = None,
    true_value: Optional[str] = None,
    false_value: Optional[str] = None,
):
    fields = list(fields)

    mapping = {k: i for i, k in enumerate(fields)}
    mask = []

    json = list(json) if json is not None else None

    for k in fields:
        if json and k in json:
            mask.append(JSON)
        else:
            mask.append(DEFAULT)

    serializer = CSVSerializer(
        plural_separator=plural_separator,
        none_value=none_value,
        true_value=true_value,
        false_value=false_value,
    )

    class Record(namedtuple(name, fields, defaults=defaults)):
        def __getitem__(self, key):
            if isinstance(key, str):
                idx = mapping.get(key)

                if idx is None:
                    raise KeyError

                return super().__getitem__(idx)

            return super().__getitem__(key)

        def get(self, key, default=None):
            try:
                return self.__getitem__(key)
            except (IndexError, KeyError):
                return default

        # NOTE: mind shadowing
        def as_csv_row(
            self,
            plural_separator=plural_separator,
            none_value=none_value,
            true_value=true_value,
            false_value=false_value,
        ):
            row = list(
                serializer(
                    v,
                    plural_separator=plural_separator,
                    none_value=none_value,
                    true_value=true_value,
                    false_value=false_value,
                    as_json=mask[i] == JSON,
                )
                for i, v in enumerate(self)
            )

            return row

        # NOTE: mind shadowing
        def as_csv_dict_row(
            self,
            plural_separator=plural_separator,
            none_value=none_value,
            true_value=true_value,
            false_value=false_value,
        ):
            row = OrderedDict(
                (
                    fields[i],
                    serializer(
                        v,
                        plural_separator=plural_separator,
                        none_value=none_value,
                        true_value=true_value,
                        false_value=false_value,
                        as_json=mask[i] == JSON,
                    ),
                )
                for i, v in enumerate(self)
            )

            return row

        def as_dict(self):
            return {fields[i]: v for i, v in enumerate(self)}

    Record.__name__ = name
    Record.is_namedrecord = True
    Record.fieldnames = fields.copy()
    Record.boolean = list(boolean) if boolean is not None else None
    Record.plural = list(plural) if plural is not None else None
    Record.json = json

    return Record


TABULAR_RECORD_SERIALIZER = CSVSerializer()
TABULAR_FIELDS = {}


def tabular_field(
    *,
    plural_separator: Optional[str] = None,
    none_value: Optional[str] = None,
    true_value: Optional[str] = None,
    false_value: Optional[str] = None,
    stringify_everything: Optional[bool] = None,
    as_json: Optional[bool] = None,
    **field_kwargs
):
    f = field(**field_kwargs)

    f_serialization_options = {}

    if plural_separator is not None:
        f_serialization_options["plural_separator"] = plural_separator

    if none_value is not None:
        f_serialization_options["none_value"] = none_value

    if true_value is not None:
        f_serialization_options["true_value"] = true_value

    if false_value is not None:
        f_serialization_options["false_value"] = false_value

    if stringify_everything is not None:
        f_serialization_options["stringify_everything"] = stringify_everything

    if as_json is not None and as_json:
        f_serialization_options["as_json"] = as_json

    if f_serialization_options:
        TABULAR_FIELDS[f] = f_serialization_options

    return f


NoneType = type(None)
PluralTypes = (list, set, tuple)


# TODO: deal with straightforward unions (Literal notably)
def parse(
    string,
    t,
    plural_separator: str = "|",
    none_value: str = "",
    true_value: str = "true",
    as_json: bool = False,
):
    if as_json:
        return json.loads(string)

    if t is str:
        return string

    if t is bool:
        return string == true_value

    if t is int:
        return int(string)

    if t is float:
        return float(string)

    origin = get_origin(t)

    if origin is None:
        raise NotImplementedError("cannot parse because type origin is unknown")

    args = get_args(t)

    if origin in PluralTypes:
        values = (
            parse(s, args[0], none_value=none_value, true_value=true_value)
            for s in string.split(plural_separator)
        )

        return origin(values)

    if origin is Union:
        if len(args) > 2 or args[1] is not NoneType:
            raise NotImplementedError(
                "cannot parse arbitrary Union except for Optional"
            )

        if string == none_value:
            return None

        return parse(
            string,
            args[0],
            plural_separator=plural_separator,
            none_value=none_value,
            true_value=true_value,
        )

    raise NotImplementedError("cannot parse arbitrary types")


def _cached_fields(cls):
    fs = cls._cached_fields

    if fs is None:
        fs = fields(cls)
        cls._cached_fields = fs

    return fs


class TabularRecord(object):
    _cached_fields = None
    _serializer_options = {
        "plural_separator": "|",
        "none_value": "",
        "true_value": "true",
        "false_value": "false",
        "stringify_everything": True,
    }

    @classmethod
    def fieldnames(cls, prefix: str = ""):
        names = []

        for f in _cached_fields(cls):
            if is_tabular_record_class(f.type):
                names.extend(f.type.fieldnames(prefix=prefix + f.name + "_"))
            else:
                names.append(prefix + f.name)

        return names

    @classmethod
    def parse(cls, row, _offset=0):
        parsed = []
        fs = _cached_fields(cls)

        options = cls._serializer_options

        i = _offset

        while i < len(row):
            v = row[i]
            f = fs[i - _offset]

            f_options = {**options, **TABULAR_FIELDS.get(f, {})}

            if is_tabular_record_class(f.type):
                i, sub_record = f.type.parse(row, _offset=i)
                parsed.append(sub_record)
            else:
                parsed.append(
                    parse(
                        v,
                        f.type,
                        plural_separator=f_options["plural_separator"],
                        none_value=f_options["none_value"],
                        true_value=f_options["true_value"],
                        as_json=f_options.get("as_json", False),
                    )
                )
                i += 1

        try:
            record = cls(*parsed)
        except TypeError:
            raise TypeError(
                "mismatch between length of rows and number of properties held by tabular record"
            )

        if _offset > 0:
            return i, record

        return record

    def as_csv_row(self):
        row = []

        options = self._serializer_options

        for f in _cached_fields(self):
            f_options = {**options, **TABULAR_FIELDS.get(f, {})}

            if is_tabular_record_class(f.type):
                row.extend(getattr(self, f.name).as_csv_row())
            else:
                row.append(
                    TABULAR_RECORD_SERIALIZER(getattr(self, f.name), **f_options)
                )

        return row

    def as_csv_dict_row(self):
        row = {}

        options = self._serializer_options

        for f in _cached_fields(self):
            f_options = {**options, **TABULAR_FIELDS.get(f, {})}

            if is_tabular_record_class(f.type):
                data = getattr(self, f.name).as_csv_dict_row()

                for n, v in data.items():
                    row[f.name + "_" + n] = v
            else:
                row[f.name] = TABULAR_RECORD_SERIALIZER(
                    getattr(self, f.name), **f_options
                )

        return row


def coerce_row(row, consume=False):
    as_csv_row = getattr(row, "as_csv_row", None)

    if callable(as_csv_row):
        row = as_csv_row()

    return list(row) if consume else row


def is_tabular_record_class(cls) -> bool:
    try:
        return issubclass(cls, TabularRecord)
    except TypeError:
        return False


def tabular_fields(cls) -> List[Field]:
    fs = []

    for f in _cached_fields(cls):
        if is_tabular_record_class(f.type):
            fs.extend(tabular_fields(f.type))
        else:
            fs.append(f)

    return fs


def coerce_fieldnames(cls):
    if is_tabular_record_class(cls):
        return cls.fieldnames()

    if getattr(cls, "is_namedrecord", False):
        return cls.fieldnames

    return cls



# =============================================================================
# Casanova Reverse Reader
# =============================================================================
#
# A reader reading the file backwards in order to read last lines in constant
# time. This is sometimes useful to be able to resume some computations
# where they were left off.
#
from casanova.reader import Reader
from casanova.exceptions import MissingColumnError


class Batch(object):
    __slots__ = ("value", "finished", "cursor", "rows")

    def __init__(self, value, finished=False, cursor=None, rows=None):
        self.value = value
        self.finished = finished
        self.cursor = cursor
        self.rows = rows or []

    def __eq__(self, other):
        return (
            self.value == other.value
            and self.finished == other.finished
            and self.cursor == other.cursor
            and self.rows == other.rows
        )

    def __iter__(self):
        return iter(self.rows)

    def collect(self, pos):
        return set(row[pos] for row in self)

    def __repr__(self):
        class_name = self.__class__.__name__

        return (
            "<%(class_name)s value=%(value)s finished=%(finished)s cursor=%(cursor)s rows=%(rows)i>"
        ) % {
            "class_name": class_name,
            "value": self.value,
            "finished": self.finished,
            "cursor": self.cursor,
            "rows": len(self.rows),
        }


class ReverseReader(Reader):
    namespace = "casanova.reverse_reader"

    def __init__(self, input_file, **kwargs):
        super().__init__(input_file, reverse=True, **kwargs)

    @staticmethod
    def last_cell(input_file, column, **kwargs):
        with ReverseReader(input_file, **kwargs) as reader:
            if reader.empty:
                return None

            return next(reader.cells(column))

    @staticmethod
    def last_batch(input_file, batch_value, batch_cursor, end_symbol, **kwargs):
        with ReverseReader(input_file, **kwargs) as reader:
            if reader.empty:
                return None

            batch = None

            if batch_value not in reader.headers:
                raise MissingColumnError(batch_value)

            if batch_cursor not in reader.headers:
                raise MissingColumnError(batch_cursor)

            batch_value_pos = reader.headers[batch_value]
            batch_cursor_pos = reader.headers[batch_cursor]

            for row in reader:
                value = row[batch_value_pos]
                cursor = row[batch_cursor_pos]

                if batch is None:
                    batch = Batch(value)

                if value != batch.value:
                    return batch

                if cursor == end_symbol:
                    batch.finished = True
                    return batch

                if cursor:
                    batch.cursor = cursor
                    return batch

                batch.rows.append(row)

            return batch



from typing import Optional, Iterable, Mapping, Any

from json import dumps
from datetime import date, datetime, time

NUMBERS = (int, float)
DATES = (date, datetime, time)
LISTS = (list, set, frozenset, tuple)


class CSVSerializer(object):
    def __init__(
        self,
        plural_separator: Optional[str] = None,
        none_value: Optional[str] = None,
        true_value: Optional[str] = None,
        false_value: Optional[str] = None,
        stringify_everything: Optional[bool] = None,
    ):
        self.plural_separator = plural_separator or "|"
        self.none_value = none_value or ""
        self.true_value = true_value or "true"
        self.false_value = false_value or "false"
        self.stringify_everything = (
            stringify_everything if stringify_everything is not None else True
        )

    def __call__(
        self,
        value,
        plural_separator: Optional[str] = None,
        none_value: Optional[str] = None,
        true_value: Optional[str] = None,
        false_value: Optional[str] = None,
        stringify_everything: Optional[bool] = None,
        as_json: bool = False,
    ):
        if as_json:
            return dumps(value, ensure_ascii=False)

        stringify_everything = (
            stringify_everything
            if stringify_everything is not None
            else self.stringify_everything
        )

        if value is None:
            none_value = none_value if none_value is not None else self.none_value

            if not none_value and not stringify_everything:
                return None

            return none_value

        if isinstance(value, str):
            return value

        if isinstance(value, bool):
            if value:
                return true_value if true_value is not None else self.true_value

            return false_value if false_value is not None else self.false_value

        if isinstance(value, NUMBERS):
            if stringify_everything:
                return str(value)
            else:
                return value

        if isinstance(value, LISTS):
            plural_separator = (
                plural_separator
                if plural_separator is not None
                else self.plural_separator
            )
            return plural_separator.join(str(i) for i in value)

        if isinstance(value, DATES):
            return value.isoformat()

        raise NotImplementedError(
            "CSVSerializer does not support this kind of value: {}".format(
                value.__class__.__name__
            )
        )

    def serialize_row(self, row: Iterable, **kwargs):
        return [self(value, **kwargs) for value in row]

    def serialize_dict_row(
        self, row: Mapping[str, Any], fieldnames: Iterable[str], **kwargs
    ):
        return [self(row.get(field), **kwargs) for field in fieldnames]



# =============================================================================
# Casanova Exceptions
# =============================================================================
#


class CasanovaError(Exception):
    pass


class NoHeadersError(CasanovaError):
    pass


class MissingColumnError(CasanovaError):
    pass


class NthNamedColumnOutOfRangeError(MissingColumnError):
    def __init__(self, name, index):
        super().__init__("%s, %i" % (name, index))
        self.name = name
        self.index = index


class UnknownNamedColumnError(MissingColumnError):
    def __init__(self, name):
        super().__init__(name)
        self.name = name


class ColumnOutOfRangeError(MissingColumnError):
    def __init__(self, index):
        super().__init__(str(index))
        self.index = index


class NotResumableError(CasanovaError):
    pass


class ResumeError(CasanovaError):
    pass


class CorruptedIndexColumnError(CasanovaError):
    pass


class Py310NullByteWriteError(CasanovaError):
    pass


class LtPy311ByteReadError(CasanovaError):
    pass


class InvalidSelectionError(CasanovaError):
    def __init__(self, msg=None, selection=None, reason=None):
        if not msg:
            if isinstance(reason, TypeError):
                msg = str(reason)
            elif isinstance(reason, IndexError):
                msg = "index %s out of range" % str(reason)
            elif isinstance(reason, KeyError):
                msg = "unknown key %s" % str(reason)

        super().__init__(msg or str(reason))
        self.selection = selection
        self.reason = reason


class NoHTTPSupport(CasanovaError):
    pass



# =============================================================================
# Casanova Library Endpoint
# =============================================================================
#
from casanova.contiguous_range_set import ContiguousRangeSet
from casanova.defaults import set_defaults, temporary_defaults
from casanova.enricher import Enricher, ThreadSafeEnricher, BatchEnricher
from casanova.headers import Headers, RowWrapper
from casanova.namedrecord import namedrecord, TabularRecord, tabular_field
from casanova.reader import Reader, Multiplexer
from casanova.resumers import (
    Resumer,
    RowCountResumer,
    ThreadSafeResumer,
    BatchResumer,
    LastCellResumer,
    LastCellComparisonResumer,
)
from casanova.reverse_reader import ReverseReader, Batch
from casanova.serialization import CSVSerializer
from casanova.utils import CsvCellIO, CsvRowIO, CsvIO
from casanova.writer import Writer

headers = Headers
reader = Reader
enricher = Enricher
threadsafe_enricher = ThreadSafeEnricher
batch_enricher = BatchEnricher
reverse_reader = ReverseReader
writer = Writer

# Re-exporting statics
count = reader.count
last_cell = reverse_reader.last_cell
last_batch = reverse_reader.last_batch



# =============================================================================
# Casanova Utils
# =============================================================================
#
# Miscellaneous utility functions.
#
import re
import csv
import gzip
import importlib
from os import PathLike
from os.path import splitext
from io import StringIO, DEFAULT_BUFFER_SIZE
from platform import python_version_tuple
from file_read_backwards.file_read_backwards import FileReadBackwardsIterator
from collections.abc import Mapping, Iterable

from casanova.exceptions import Py310NullByteWriteError, LtPy311ByteReadError

PY_310 = python_version_tuple()[:2] == ("3", "10")
LT_PY311 = python_version_tuple()[:2] <= ("3", "10")


def py310_wrap_csv_writerow(writer):
    if not PY_310:
        return writer.writerow

    def wrapped(*args, **kwargs):
        try:
            writer.writerow(*args, **kwargs)
        except csv.Error as e:
            if str(e).lower() == "need to escape, but no escapechar set":
                raise Py310NullByteWriteError(
                    "Cannot write row containing null byte. This error only happens on python 3.10 (see https://github.com/python/cpython/issues/56387). Consider using the strip_null_bytes_on_write=True kwarg or change python version."
                )

            raise

    return wrapped


def ltpy311_csv_reader(input_file, **kwargs):
    reader = csv.reader(input_file, **kwargs)

    if not LT_PY311:
        return reader

    def wrapped():
        try:
            for item in reader:
                yield item
        except csv.Error as e:
            if "line contains nul" in str(e).lower():
                raise LtPy311ByteReadError(
                    "python < 3.11 cannot read CSV files containing null bytes. Consider using the strip_null_bytes_on_read=True kwarg or upgrade your python version."
                )

            raise

    return wrapped()


def ensure_open(p, mode="r", encoding="utf-8", newline=None):
    if not isinstance(p, (str, PathLike)):
        return p

    p = str(p)

    if p.endswith(".gz"):
        if "b" in mode:
            return gzip.open(p, mode=mode, newline=newline)

        mode += "t"
        return gzip.open(p, encoding=encoding, mode=mode, newline=newline)

    if "b" in mode:
        return open(p, mode=mode, newline=newline)

    return open(p, encoding=encoding, mode=mode, newline=newline)


def parse_module_and_function(path):
    if ":" in path:
        s = path.rsplit(":", 1)
        return s[0], s[1]

    return path, "main"


def import_function(path):
    module_name, function_name = parse_module_and_function(path)
    m = importlib.import_module(module_name)
    return getattr(m, function_name)


def infer_delimiter_or_type(file_or_path):
    if isinstance(file_or_path, PathLike):
        file_or_path = str(file_or_path)

    if hasattr(file_or_path, "name"):
        file_or_path = file_or_path.name

    if isinstance(file_or_path, str):
        rest, ext = splitext(file_or_path)

        if ext == ".gz":
            _, ext = splitext(rest)

        if ext == ".tsv" or ext == ".tab":
            return ("csv", "\t")

        if ext == ".ndjson" or ext == ".jsonl":
            return ("ndjson", None)

    return ("csv", None)


BOM_RE = re.compile(r"^\ufeff")


def suppress_BOM(string):
    return re.sub(BOM_RE, "", string)


def has_null_byte(string):
    return "\0" in string


def strip_null_bytes(string):
    return string.replace("\0", "")


def lines_without_null_bytes(iterable):
    for line in iterable:
        yield strip_null_bytes(line)


def first_cell_index_with_null_byte(row):
    for i, cell in enumerate(row):
        if has_null_byte(cell):
            return i

    return None


def strip_null_bytes_from_row(row):
    if any(has_null_byte(cell) for cell in row if isinstance(cell, str)):
        return [
            strip_null_bytes(cell) if isinstance(cell, str) else cell for cell in row
        ]

    return row


def rows_without_null_bytes(iterable):
    for row in iterable:
        yield strip_null_bytes_from_row(row)


def looks_like_url(value: str) -> bool:
    return value.startswith("http://") or value.startswith("https://")


def size_of_row_in_memory(row):
    """
    Returns the approximate amount of bytes needed to represent the given row into
    the python's program memory.

    The magic numbers are based on `sys.getsizeof`.
    """
    a = 64 + 8 * len(row)  # Size of the array
    a += sum(49 + len(cell) for cell in row)  # Size of the contained strings

    return a


def size_of_row_in_file(row):
    """
    Returns the approximate amount of bytes originally used to represent the
    given row in its CSV file. It assumes the delimiter uses only one byte.

    I also ignores quotes (-2 bytes) around escaped cells if they were
    originally present.

    I also don't think that it counts 16 bit chars correctly.
    """
    a = max(0, len(row) - 1)
    a += sum(len(cell) for cell in row)

    return a


def flatmap(item):
    if not isinstance(item, Iterable) or isinstance(item, (str, bytes)):
        yield item
    else:
        for sub_item in item:
            yield from flatmap(sub_item)


class CsvIOBase(StringIO):
    def __init__(self):
        super().__init__()
        self.writer = csv.writer(
            self, dialect=csv.unix_dialect, quoting=csv.QUOTE_MINIMAL
        )


class CsvCellIO(CsvIOBase):
    def __init__(self, value, column=None):
        super().__init__()

        if column is not None:
            self.fieldnames = [column]
            self.writer.writerow(self.fieldnames)

        self.writer.writerow([value])

        self.seek(0)


class CsvRowIO(CsvIOBase):
    def __init__(self, row, fieldnames=None):
        super().__init__()

        self.fieldnames = fieldnames

        if isinstance(row, Mapping):
            if self.fieldnames is None:
                self.fieldnames = list(row.keys())

            row = [row.get(f) for f in self.fieldnames]

        if self.fieldnames is not None:
            self.writer.writerow(self.fieldnames)

        self.writer.writerow(row)

        self.seek(0)


class CsvIO(CsvIOBase):
    def __init__(self, rows, fieldnames=None):
        super().__init__()

        self.fieldnames = fieldnames

        if fieldnames is not None:
            self.writer.writerow(fieldnames)

        for row in rows:
            self.writer.writerow(row)

        self.seek(0)


def create_csv_aware_backwards_lines_iterator(
    input_file, quotechar=None, strip_null_bytes_on_read=False
):
    encoding = input_file.encoding if hasattr(input_file, "encoding") else "utf-8"
    quotechar = quotechar if quotechar is not None else '"'

    backwards_file = ensure_open(input_file.name, mode="rb")

    backwards_iterator = FileReadBackwardsIterator(
        backwards_file, encoding, DEFAULT_BUFFER_SIZE
    )

    def correctly_escaped_backwards_iterator():
        acc = None

        for line in backwards_iterator:
            if strip_null_bytes_on_read:
                line = strip_null_bytes(line)

            if acc is not None:
                acc = line + "\n" + acc
            else:
                acc = line

            if acc.count(quotechar) % 2 == 0:
                yield acc
                acc = None

        if acc is not None:
            yield acc

    return backwards_file, correctly_escaped_backwards_iterator()



# =============================================================================
# Casanova Enricher
# =============================================================================
#
# A CSV reader/writer combo that can be used to read an input CSV file and
# easily ouput a similar CSV file while editing, adding and filtering cell_count.
#
from ebbe import with_is_last

from casanova.resumers import (
    LastCellComparisonResumer,
    Resumer,
    RowCountResumer,
    ThreadSafeResumer,
    BatchResumer,
)
from casanova.headers import Headers
from casanova.reader import Reader
from casanova.writer import Writer
from casanova.namedrecord import coerce_row, coerce_fieldnames
from casanova.defaults import DEFAULTS


class Enricher(Reader):
    __supported_resumers__ = (RowCountResumer, LastCellComparisonResumer)

    def __init__(
        self,
        input_file,
        output_file,
        no_headers=False,
        select=None,
        add=None,
        strip_null_bytes_on_write=None,
        writer_dialect=None,
        writer_delimiter=None,
        writer_quotechar=None,
        writer_escapechar=None,
        writer_quoting=None,
        writer_lineterminator=None,
        write_header=True,
        **kwargs
    ):
        # Inheritance
        super().__init__(input_file, no_headers=no_headers, **kwargs)

        if strip_null_bytes_on_write is None:
            strip_null_bytes_on_write = DEFAULTS.strip_null_bytes_on_write

        if not isinstance(strip_null_bytes_on_write, bool):
            raise TypeError('expecting a boolean as "strip_null_bytes_on_write" kwarg')

        self.strip_null_bytes_on_write = strip_null_bytes_on_write

        self.selected_indices = None
        self.output_fieldnames = self.fieldnames
        self.added_count = 0

        if select is not None:
            if no_headers:
                self.selected_indices = Headers.select_no_headers(self.row_len, select)
            else:
                self.selected_indices = self.headers.select(select)
                self.output_fieldnames = self.__filterrow(self.output_fieldnames)

        add = coerce_fieldnames(add)

        if add is not None:
            if no_headers:
                if not isinstance(add, int):
                    raise TypeError(
                        "cannot add named columns with no_headers=True. Use a number of columns instead."
                    )

                self.added_count = add
            else:
                self.output_fieldnames += add
                self.added_count = len(add)

        self.padding = [""] * self.added_count

        # Resuming?
        self.resumer = None
        can_resume = False

        if isinstance(output_file, Resumer):
            if not isinstance(output_file, self.__class__.__supported_resumers__):
                raise TypeError(
                    "%s: does not support %s!"
                    % (self.__class__.__name__, output_file.__class__.__name__)
                )

            self.resumer = output_file

            can_resume = self.resumer.can_resume()

            if can_resume:
                # NOTE: how about null bytes
                self.resumer.get_insights_from_output(
                    self,
                    no_headers=no_headers,
                    dialect=writer_dialect,
                    quotechar=writer_quotechar,
                    delimiter=writer_delimiter,
                )

                if hasattr(self.resumer, "resume"):
                    self.resumer.resume(self)

            output_file = self.resumer.open_output_file()

            if hasattr(self.resumer, "filter"):
                self.row_filter = self.resumer.filter_row
            else:
                self.prelude_rows = self.resumer

        # Instantiating writer
        self.writer = Writer(
            output_file,
            fieldnames=self.output_fieldnames,
            strip_null_bytes_on_write=strip_null_bytes_on_write,
            dialect=writer_dialect,
            delimiter=writer_delimiter,
            quotechar=writer_quotechar,
            escapechar=writer_escapechar,
            quoting=writer_quoting,
            lineterminator=writer_lineterminator,
            write_header=not can_resume and write_header,
        )

    def __repr__(self):
        return "<%s>" % self.__class__.__name__

    @property
    def output_headers(self):
        return self.writer.headers

    @property
    def should_write_header(self):
        return self.writer.should_write_header

    def __filterrow(self, row):
        row = coerce_row(row)

        if self.selected_indices is not None:
            row = [row[i] for i in self.selected_indices]

        return row

    def __formatrow(self, row, add=None, *addenda):
        # We expect additions
        if self.added_count > 0:
            if add is None:
                add = self.padding
            else:
                add = coerce_row(add, consume=True)

                for addition in addenda:
                    addition = coerce_row(addition)
                    add.extend(addition)

                if len(add) != self.added_count:
                    raise TypeError(
                        "casanova.enricher.writerow: expected %i additional cells but got %i."
                        % (self.added_count, len(add))
                    )

            formatted_row = self.__filterrow(row) + add

        # We don't expect additions
        else:
            if add is not None:
                raise TypeError("casanova.enricher.writerow: expected no additions.")

            formatted_row = self.__filterrow(row)

        return formatted_row

    def writeheader(self):
        self.writer.writeheader()

    def writerow(self, row, add=None, *addenda):
        self.writer.writerow(self.__formatrow(row, add, *addenda))

    def writebatch(self, row, addenda):
        row = self.__filterrow(row)

        for addendum in addenda:
            addendum = coerce_row(addendum)

            if len(addendum) != self.added_count:
                raise TypeError(
                    "casanova.enricher.writebatch: expected %i additional cells but got %i."
                    % (self.added_count, len(addendum))
                )

            self.writer.writerow(row + addendum)


class ThreadSafeEnricher(Enricher):
    __supported_resumers__ = (ThreadSafeResumer,)

    def __init__(
        self, input_file, output_file, add=None, index_column="index", **kwargs
    ):
        self.index_column = index_column

        # Prepending index column to output
        add = [index_column] + list(coerce_fieldnames(add or []))

        # Inheritance
        super().__init__(input_file, output_file, add=add, **kwargs)

    def __iter__(self):
        return self.enumerate()

    def cells(self, column, with_rows=False):
        return self.enumerate_cells(column, with_rows=with_rows)

    def records(self, *shape, with_rows=False):
        return self.enumerate_records(*shape, with_rows=with_rows)

    def writerow(self, index, row, add=None, *addenda):
        if add is None:
            super().writerow(row, [index] + self.padding[:-1])
        else:
            super().writerow(row, [index], add, *addenda)


class BatchEnricher(Enricher):
    __supported_resumers__ = (BatchResumer,)

    def __init__(
        self,
        input_file,
        output_file,
        add=None,
        cursor_column="cursor",
        end_symbol="end",
        **kwargs
    ):
        self.cursor_column = cursor_column
        self.end_symbol = end_symbol

        # Prepending cursor column to output
        add = [cursor_column] + list(coerce_fieldnames(add or []))

        # Inheritance
        super().__init__(input_file, output_file, add=add, **kwargs)

    def writebatch(self, row, addenda, cursor=None):
        if cursor is None:
            cursor = self.end_symbol

        for is_last, addendum in with_is_last(addenda):
            self.writerow(row, [cursor if is_last else None], addendum)



from typing import Optional, List

import re
import sys
import json
import math
import random
import statistics
from types import GeneratorType
from os.path import join
from urllib.parse import urlsplit, urljoin
from multiprocessing import Pool as MultiProcessPool
from dataclasses import dataclass
from collections import Counter, defaultdict, deque, OrderedDict
from collections.abc import Mapping, Iterable

from casanova import Reader, Enricher, CSVSerializer, RowWrapper, Headers, Writer
from casanova.utils import import_function, flatmap


@dataclass
class InitializerOptions:
    code: str
    module: bool
    row_len: int
    args: List[str]
    init_codes: List[str]
    before_codes: List[str]
    after_codes: List[str]
    fieldnames: Optional[List[str]] = None
    selected_indices: Optional[List[int]] = None


# NOTE: just a thin wrapper to make sure we catch KeyboardInterrupt in
# child processes gracefully.
class WorkerWrapper(object):
    __slots__ = ("fn",)

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args, **kwargs):
        try:
            return self.fn(*args, **kwargs)
        except KeyboardInterrupt:
            sys.exit(1)


class SingleProcessPool(object):
    def imap(self, worker, tasks, chunksize=1):
        for t in tasks:
            yield worker(t)

    def imap_unordered(self, *args, **kwargs):
        yield from self.imap(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return


def get_pool(n: int, options: InitializerOptions):
    initargs = (options,)

    if n < 2:
        multiprocessed_initializer(*initargs)
        return SingleProcessPool()

    return MultiProcessPool(
        n, initializer=multiprocessed_initializer, initargs=initargs
    )


def get_csv_serializer(cli_args):
    return CSVSerializer(
        plural_separator=cli_args.plural_separator,
        none_value=cli_args.none_value,
        true_value=cli_args.true_value,
        false_value=cli_args.false_value,
    )


# Global multiprocessing variables
CODE = None
FUNCTION = None
ARGS = None
SELECTION = None
BEFORE_CODES = []
AFTER_CODES = []
EVALUATION_CONTEXT = {}
ROW = None

EVALUATION_CONTEXT_LIB = {
    # lib
    "join": join,
    "math": math,
    "mean": statistics.mean,
    "median": statistics.median,
    "random": random,
    "re": re,
    "urljoin": urljoin,
    "urlsplit": urlsplit,
    # classes
    "Counter": Counter,
    "defaultdict": defaultdict,
    "deque": deque,
}


def initialize_evaluation_context():
    global EVALUATION_CONTEXT

    EVALUATION_CONTEXT = {
        **EVALUATION_CONTEXT_LIB,
        # state
        "fieldnames": None,
        "headers": None,
        "index": 0,
        "row": None,
        "cell": None,
        "cells": None,
    }


def multiprocessed_initializer(options: InitializerOptions):
    global CODE
    global FUNCTION
    global ARGS
    global BEFORE_CODES
    global AFTER_CODES
    global ROW
    global SELECTION

    # Reset in case of multiple execution from same process
    CODE = None
    FUNCTION = None
    ARGS = None
    SELECTION = None
    BEFORE_CODES = []
    AFTER_CODES = []
    ROW = None
    initialize_evaluation_context()

    if options.module:
        FUNCTION = import_function(options.code)
        ARGS = options.args
    else:
        CODE = options.code
        BEFORE_CODES = options.before_codes
        AFTER_CODES = options.after_codes

    if options.selected_indices is not None:
        SELECTION = options.selected_indices

    if options.fieldnames is not None:
        EVALUATION_CONTEXT["fieldnames"] = options.fieldnames
        EVALUATION_CONTEXT["headers"] = Headers(options.fieldnames)
        headers = EVALUATION_CONTEXT["headers"]
    else:
        headers = Headers(range(options.row_len))

    for init_code in options.init_codes:
        exec(init_code, None, EVALUATION_CONTEXT)

    EVALUATION_CONTEXT["row"] = RowWrapper(headers, None)
    ROW = EVALUATION_CONTEXT["row"]


def select(row):
    if SELECTION is None:
        return

    cells = [row[i] for i in SELECTION]
    EVALUATION_CONTEXT["cells"] = cells
    EVALUATION_CONTEXT["cell"] = cells[0]


def multiprocessed_worker_using_eval(payload):
    global EVALUATION_CONTEXT

    i, row = payload
    EVALUATION_CONTEXT["index"] = i
    ROW._replace(row)

    select(row)

    try:
        for before_code in BEFORE_CODES:
            exec(before_code, EVALUATION_CONTEXT, None)

        value = eval(CODE, EVALUATION_CONTEXT, None)

        for after_code in AFTER_CODES:
            exec(after_code, EVALUATION_CONTEXT, None)

        return None, i, value
    except Exception as e:
        return e, i, None


def collect_args(i, row):
    for arg_name in ARGS:
        if arg_name == "row":
            yield ROW
        elif arg_name == "index":
            yield i
        elif arg_name == "fieldnames":
            yield EVALUATION_CONTEXT["fieldnames"]
        elif arg_name == "headers":
            yield EVALUATION_CONTEXT["headers"]
        elif arg_name == "cell":
            # NOTE: we know SELECTION is relevant because it's validated by CLI
            yield row[SELECTION[0]]
        elif arg_name == "cells":
            # NOTE: we know SELECTION is relevant because it's validated by CLI
            for idx in SELECTION:
                yield row[idx]
        else:
            raise TypeError("unknown arg_name: %s" % arg_name)


def multiprocessed_worker_using_function(payload):
    i, row = payload
    ROW._replace(row)

    args = tuple(collect_args(i, row))

    try:
        value = FUNCTION(*args)

        # NOTE: consuming generators
        if isinstance(value, GeneratorType):
            value = list(value)

        return None, i, value
    except Exception as e:
        return e, i, None


# TODO: go to minet for progress bar and rich?
# TODO: write proper cli documentation
def mp_iteration(cli_args, reader: Reader):
    worker = (
        multiprocessed_worker_using_eval
        if not cli_args.module
        else multiprocessed_worker_using_function
    )

    if cli_args.processes > 1:
        worker = WorkerWrapper(worker)

    selected_indices = None

    if cli_args.select:
        if reader.headers is not None:
            selected_indices = reader.headers.select(cli_args.select)
        else:
            selected_indices = Headers.select_no_headers(cli_args.select)

    init_options = InitializerOptions(
        code=cli_args.code,
        module=cli_args.module,
        args=cli_args.args,
        init_codes=cli_args.init,
        before_codes=cli_args.before,
        after_codes=cli_args.after,
        row_len=reader.row_len,
        fieldnames=reader.fieldnames,
        selected_indices=selected_indices,
    )

    with get_pool(cli_args.processes, init_options) as pool:
        # NOTE: we keep track of rows being worked on from the main process
        # to avoid serializing them back with worker result.
        worked_rows = {}

        def payloads():
            for t in reader.enumerate():
                worked_rows[t[0]] = t[1]
                yield t

        mapper = pool.imap if not cli_args.unordered else pool.imap_unordered

        for exc, i, result in mapper(worker, payloads(), chunksize=cli_args.chunk_size):
            row = worked_rows.pop(i)

            if exc is not None:
                if cli_args.ignore_errors:
                    result = None
                else:
                    raise exc

            yield i, row, result


def map_action(cli_args, output_file):
    serialize = get_csv_serializer(cli_args)

    with Enricher(
        cli_args.file,
        output_file,
        add=[cli_args.new_column],
        delimiter=cli_args.delimiter,
    ) as enricher:
        for _, row, result in mp_iteration(cli_args, enricher):
            enricher.writerow(row, [serialize(result)])


def flatmap_action(cli_args, output_file):
    serialize = get_csv_serializer(cli_args)

    with Enricher(
        cli_args.file,
        output_file,
        add=[cli_args.new_column],
        delimiter=cli_args.delimiter,
    ) as enricher:
        for _, row, result in mp_iteration(cli_args, enricher):
            for value in flatmap(result):
                enricher.writerow(row, [serialize(value)])


def filter_action(cli_args, output_file):
    with Enricher(cli_args.file, output_file, delimiter=cli_args.delimiter) as enricher:
        for _, row, result in mp_iteration(cli_args, enricher):
            if result:
                enricher.writerow(row)


def map_reduce_action(cli_args, output_file):
    acc_fn = None

    if cli_args.module:
        acc_fn = import_function(cli_args.accumulator)

    with Reader(
        cli_args.file,
        delimiter=cli_args.delimiter,
    ) as enricher:
        acc_context = EVALUATION_CONTEXT_LIB.copy()

        acc = None
        initialized = False

        if cli_args.init_value is not None:
            initialized = True
            acc = eval(cli_args.init_value, acc_context, None)

        acc_context["acc"] = acc

        for _, row, result in mp_iteration(cli_args, enricher):
            if not initialized:
                acc_context["acc"] = result
                initialized = True
                continue

            if acc_fn is None:
                acc_context["current"] = result
                acc_context["acc"] = eval(cli_args.accumulator, acc_context, None)
            else:
                acc_context["acc"] = acc_fn(acc_context["acc"], result)

        final_result = acc_context["acc"]

        if cli_args.json:
            json.dump(
                final_result,
                output_file,
                indent=2 if cli_args.pretty else None,
                ensure_ascii=False,
            )
            print(file=output_file)
        elif cli_args.csv:
            writer = Writer(output_file)

            serializer = get_csv_serializer(cli_args)
            fieldnames = ["value"]

            if isinstance(final_result, Mapping):
                fieldnames = list(final_result.keys())
                writer.writerow(fieldnames)
                writer.writerow(serializer.serialize_dict_row(final_result, fieldnames))
            elif isinstance(final_result, Iterable) and not isinstance(
                final_result, (bytes, str)
            ):
                writer.writerow(serializer.serialize_row(final_result))
            else:
                writer.writerow(fieldnames)
                writer.writerow(serializer(final_result))

        else:
            print(final_result, file=output_file)


class GroupWrapper:
    __slots__ = ("__name", "__rows", "__wrapper")

    def __init__(self, fieldnames):
        self.__wrapper = RowWrapper(Headers(fieldnames), range(len(fieldnames)))

    def _replace(self, name, rows):
        self.__name = name
        self.__rows = rows

    @property
    def name(self):
        return self.__name

    def __len__(self):
        return len(self.__rows)

    def __iter__(self):
        for row in self.__rows:
            self.__wrapper._replace(row)
            yield self.__wrapper


def groupby_action(cli_args, output_file):
    agg_fn = None

    if cli_args.module:
        agg_fn = import_function(cli_args.aggregator)

    with Reader(
        cli_args.file,
        delimiter=cli_args.delimiter,
    ) as enricher:
        # NOTE: using an ordered dict to guarantee stability for all python versions
        groups = OrderedDict()

        # Grouping
        for _, row, result in mp_iteration(cli_args, enricher):
            l = groups.get(result)

            if l is None:
                l = [row]
                groups[result] = l
            else:
                l.append(row)

        # Aggregating
        agg_context = EVALUATION_CONTEXT_LIB.copy()
        header_emitted = False

        writer = Writer(output_file)
        fieldnames = ["group"]
        mapping_fieldnames = None
        serializer = get_csv_serializer(cli_args)

        if cli_args.fieldnames is not None:
            mapping_fieldnames = cli_args.fieldnames
            fieldnames += cli_args.fieldnames
            header_emitted = True
            writer.writerow(fieldnames)

        group_wrapper = GroupWrapper(enricher.fieldnames)

        for name, rows in groups.items():
            group_wrapper._replace(name, rows)

            if agg_fn is not None:
                result = agg_fn(group_wrapper)
            else:
                agg_context["group"] = group_wrapper
                result = eval(cli_args.aggregator, agg_context, None)

            name = serializer(name)

            if isinstance(result, Mapping):
                if not header_emitted:
                    mapping_fieldnames = list(result.keys())
                    fieldnames += mapping_fieldnames
                    writer.writerow(fieldnames)
                    header_emitted = True

                writer.writerow(
                    [name] + serializer.serialize_dict_row(result, mapping_fieldnames)
                )
            elif isinstance(result, Iterable) and not isinstance(result, (bytes, str)):
                writer.writerow([name] + serializer.serialize_row(result))
            else:
                if not header_emitted:
                    writer.writerow(fieldnames + ["value"])
                    header_emitted = True
                writer.writerow([name, serializer(result)])


def reverse_action(cli_args, output_file):
    with Enricher(
        cli_args.file, output_file, delimiter=cli_args.delimiter, reverse=True
    ) as enricher:
        for row in enricher:
            enricher.writerow(row)



from setuptools import setup, find_packages

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="casanova",
    version="1.8.0-a1",
    description="Specialized & performant CSV readers, writers and enrichers for python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/medialab/casanova",
    license="MIT",
    author="Guillaume Plique",
    author_email="kropotkinepiotr@gmail.com",
    keywords="csv",
    python_requires=">=3.7",
    packages=find_packages(exclude=["benchmark", "test"]),
    package_data={"docs": ["README.md"]},
    install_requires=["ebbe>=1.11.0,<2", "file-read-backwards>=2.0.0,<3"],
    extras_require={
        ":python_version<'3.10'": ["typing_extensions"],
        "http": ["urllib3", "certifi"],
    },
    entry_points={
        "console_scripts": [
            "casanova=casanova.__main__:main",
            "casa=casanova.__main__:main",
        ]
    },
    zip_safe=True,
)



import json


class writer(object):
    def __init__(self, f):
        self.__file = f

    def writerow(self, row):
        self.__file.write(json.dumps(row, ensure_ascii=False, allow_nan=False))
        self.__file.write("\n")


class reader(object):
    def __init__(self, f):
        self.__file = f

    def __next__(self):
        line = ""

        while not line:
            line = next(self.__file).strip()

        return json.loads(line)



# =============================================================================
# Casanova Writer
# =============================================================================
#
# A CSV writer that is only really useful if you intend to resume its operation
# somehow
#
import csv

from casanova.defaults import DEFAULTS
from casanova.resumers import Resumer, LastCellResumer
from casanova.namedrecord import coerce_row, coerce_fieldnames
from casanova.reader import Headers
from casanova.utils import py310_wrap_csv_writerow, strip_null_bytes_from_row


class Writer(object):
    __supported_resumers__ = (LastCellResumer,)

    def __init__(
        self,
        output_file,
        fieldnames=None,
        strip_null_bytes_on_write=None,
        dialect=None,
        delimiter=None,
        quotechar=None,
        quoting=None,
        escapechar=None,
        lineterminator=None,
        write_header=True,
    ):
        if strip_null_bytes_on_write is None:
            strip_null_bytes_on_write = DEFAULTS.strip_null_bytes_on_write

        if not isinstance(strip_null_bytes_on_write, bool):
            raise TypeError('expecting a boolean as "strip_null_bytes_on_write" kwarg')

        self.strip_null_bytes_on_write = strip_null_bytes_on_write

        fieldnames = coerce_fieldnames(fieldnames)

        self.fieldnames = fieldnames
        self.headers = Headers(fieldnames) if fieldnames is not None else None
        self.no_headers = fieldnames is None

        can_resume = False

        if isinstance(output_file, Resumer):
            if self.fieldnames is None:
                raise NotImplementedError

            resumer = output_file

            if not isinstance(output_file, self.__class__.__supported_resumers__):
                raise TypeError(
                    "%s: does not support %s!"
                    % (self.__class__.__name__, output_file.__class__.__name__)
                )

            can_resume = resumer.can_resume()

            if can_resume:
                # NOTE: how about null bytes
                resumer.get_insights_from_output(
                    self,
                    no_headers=self.no_headers,
                    dialect=dialect,
                    quotechar=quotechar,
                    delimiter=delimiter,
                )

            output_file = resumer.open_output_file()

        # Instantiating writer
        self.dialect = dialect
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.quoting = quoting
        self.lineterminator = lineterminator

        writer_kwargs = {}

        if dialect is not None:
            writer_kwargs["dialect"] = dialect

        if delimiter is not None:
            writer_kwargs["delimiter"] = delimiter

        if quotechar is not None:
            writer_kwargs["quotechar"] = quotechar

        if escapechar is not None:
            writer_kwargs["escapechar"] = escapechar

        if quoting is not None:
            writer_kwargs["quoting"] = quoting

        if lineterminator is not None:
            writer_kwargs["lineterminator"] = lineterminator

        self.__writer = csv.writer(output_file, **writer_kwargs)

        if not strip_null_bytes_on_write:
            self.__writerow = py310_wrap_csv_writerow(self.__writer)
        else:
            self.__writerow = lambda row: self.__writer.writerow(
                strip_null_bytes_from_row(row)
            )

        self.should_write_header = not can_resume and self.fieldnames is not None

        if self.should_write_header and write_header:
            self.writeheader()

    def writerow(self, row):
        row = coerce_row(row)
        self.__writerow(row)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

    def writeheader(self):
        if self.fieldnames is None:
            raise TypeError("cannot write header if fieldnames were not provided")

        self.should_write_header = False

        row = self.fieldnames

        if self.strip_null_bytes_on_write:
            row = strip_null_bytes_from_row(row)

        self.writerow(row)
